﻿// <copyright file="Program.cs" company="Wired Digital Ltd">
// Copyright (c) 2015 All Right Reserved, http://umbracofreelancer.uk
//
// All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY 
// KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// </copyright>
// <author>David Peck</author>
// <email>davuid@wired-digital.co.uk</email>
// <date>2015-11-15</date>
// <summary>Synchronises a local directory with a Microsoft Windows Azure storage container</summary>

using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using Microsoft.WindowsAzure.Storage;
using Microsoft.WindowsAzure.Storage.Auth;
using Microsoft.WindowsAzure.Storage.Blob;
using MimeTypes;

namespace AzureStorageSync
{
    /// <summary>
    ///     Synchronises a local directory with a  Microsoft Windows Azure storage container
    /// </summary>
    /// <remarks>Settings can be provided as console arguments or App.config AppSettings</remarks>
    internal class Program
    {
        private const string SyntaxHelp =
            "AzureStorageSync /acc:account_name /key:key_value /path:local_path /container:container_name /force_properties=1";

        //The maximum cache age is 10 days for all files. 
        //You may want to modify this value or adjust it for different file types/ paths
        private static TimeSpan _maxAge = TimeSpan.FromDays(10);

        private static void Main(string[] args)
        {
            if (!ParseArgs(args))
                return;

            //Specify storage credentials.
            var credentials = new StorageCredentials(AccountName, KeyValue);

            //Create a reference to your storage account, passing in your credentials.
            var storageAccount = new CloudStorageAccount(credentials, true);

            //Create a new client object, which will provide access to Blob service resources.
            var blobClient = storageAccount.CreateCloudBlobClient();

            //Create a new container.
            var container = blobClient.GetContainerReference(ContainerName);
            container.CreateIfNotExists();

            //Upload all new or updated assets
            var uploaded = UploadDirectory(LocalPath, container);
            Console.WriteLine("Uploaded {0} blob(s)", uploaded);

            //Remove anything in storage which isn't also local
            var deleted = RemoveOrphanedBlobs(container);
            Console.WriteLine("Deleted {0} blob(s)", deleted);
        }

        /// <summary>
        ///     Requestion file upload for all files within this directory and recursively call for all sub directories
        /// </summary>
        /// <param name="directoryPath">The current director to upload</param>
        /// <param name="container">The Azure container to upload to</param>
        /// <returns>The number of files uploaded</returns>
        private static int UploadDirectory(string directoryPath, CloudBlobContainer container)
        {
            //Upload all files in this directory
            var counter = Directory.GetFiles(directoryPath).Count(filePath => UploadFile(filePath, container));

            //Recursive call for all sub directories
            counter += Directory.GetDirectories(directoryPath).Sum(filePath => UploadDirectory(filePath, container));

            return counter;
        }

        /// <summary>
        ///     Upload a file, if the file doesn't exist, or is different to the local file
        /// </summary>
        /// <param name="filePath">The file path to upload</param>
        /// <param name="container">The container to upload to</param>
        /// <returns>If the upload was required</returns>
        private static bool UploadFile(string filePath, CloudBlobContainer container)
        {
            //Work out what the blob should be called as a Uri relative to the local directory
            var fileUri = new Uri(filePath);
            var blobName = new Uri(LocalPath).MakeRelativeUri(fileUri).OriginalString;

            //Store if we need to do anything with this file.
            var uploadRequired = true;

            //Get a local reference to the blob (not checked with server)
            ICloudBlob cloudBlob = container.GetBlockBlobReference(blobName);

            //Is this a new file?
            if (cloudBlob.Exists())
            {
                //Replace cloud blob with a version fetched from the server
                cloudBlob = container.GetBlobReferenceFromServer(cloudBlob.Name);

                //Check if the file has changed based on MD5 hash
                var md5 = GetMd5(filePath);
                uploadRequired = !md5.Equals(cloudBlob.Properties.ContentMD5);
            }

            if (uploadRequired)
            {

                //Upload the file
                cloudBlob.UploadFromFile(filePath, FileMode.Open);
            }

            if (uploadRequired || !ForceBlobProperties)
            {
                //Set the correct blob properties (cache and mime type)
                UpdateBlobProperties(cloudBlob);
            }

            return uploadRequired;
        }

        /// <summary>
        ///     Set the properties of the blob, if different to those specified
        /// </summary>
        /// <param name="cloudBlob">The blob to update</param>
        private static void UpdateBlobProperties(ICloudBlob cloudBlob)
        {
            //This is the value that will be used in the cache-control header
            var cacheControlValue = "public, max-age=" + _maxAge.TotalSeconds;

            var modified = false;

            //Set the correct mime type, if different
            var mimeType = GetMimeType(cloudBlob.Uri);
            if (cloudBlob.Properties.ContentType != mimeType)
            {
                cloudBlob.Properties.ContentType = mimeType;
                modified = true;
            }

            //Set the cache control, if different
            if (cloudBlob.Properties.CacheControl != cacheControlValue)
            {
                cloudBlob.Properties.CacheControl = cacheControlValue;
                modified = true;
            }

            //If anything changed then tell the server
            if (modified)
                cloudBlob.SetProperties();
        }

        /// <summary>
        ///     Check for deleted local files
        /// </summary>
        /// <param name="container">The container we're uploading to</param>
        /// <returns>The number of files deleted</returns>
        private static int RemoveOrphanedBlobs(CloudBlobContainer container)
        {
            var count = 0;

            //For all blobs on the server
            foreach (var blobRef in container.ListBlobs(useFlatBlobListing: true).Cast<CloudBlockBlob>())
            {
                var localFileInfo = new FileInfo(Path.Combine(LocalPath, blobRef.Name));

                //check if a local file exists
                if (localFileInfo.Exists)
                {
                    //Do a secondary case sensitiveness match, as blobs are case sensitive and local file paths aren't
                    var storagePathParts = blobRef.Name
                            .Split(new[] { '/' })
                            .Reverse()
                            .ToArray();
                    var localPathParts = GetProperFilePathCapitalization(localFileInfo.FullName)
                            .Split(new[] { '\\' })
                            .Reverse();
                    if (localPathParts
                            .Take(storagePathParts.Length)
                            .SequenceEqual(storagePathParts, StringComparer.InvariantCulture))
                        continue;
                }

                //Delete anything that doesn't exist locally
                var blockBlob = container.GetBlockBlobReference(blobRef.Name);
                blockBlob.Delete();
                count++;
            }

            return count;
        }

        #region Support methods

        /// <summary>
        ///     Find the arguments specified and populate the relevant properties
        /// </summary>
        private static bool ParseArgs(IEnumerable<string> args)
        {
            string localPath = null;
            string containerName = null;
            string accountName = null;
            string keyValue = null;
            bool? forceProperties = null;
            try
            {
                foreach (var arg in args)
                {
                    var argParts = arg.Split(new[] { ':' }, 2);
                    if (argParts.Length != 2 || !argParts[0].StartsWith("/"))
                        throw new FormatException(string.Format("Unknown argument: {0}.", arg));

                    var ident = argParts[0].Substring(1).ToLower();
                    switch (ident)
                    {
                        case "acc":
                            accountName = argParts[1];
                            break;
                        case "key":
                            keyValue = argParts[1];
                            break;
                        case "path":
                            localPath = argParts[1];
                            break;
                        case "container":
                            containerName = argParts[1];
                            break;
                        case "force_properties":
                            forceProperties = argParts[1] == "1";
                            break;
                        default:
                            throw new FormatException(string.Format("Unknown argument: {0}.", arg));
                    }
                }

                LocalPath = localPath;
                ContainerName = containerName;
                AccountName = accountName;
                KeyValue = keyValue;
                if (forceProperties.HasValue)
                    ForceBlobProperties = forceProperties.Value;

                if (string.IsNullOrWhiteSpace(LocalPath))
                    throw new FormatException("The local path has not been supplied.");
                if (string.IsNullOrWhiteSpace(ContainerName))
                    throw new FormatException("The container name has not been supplied.");
                if (string.IsNullOrWhiteSpace(AccountName))
                    throw new FormatException("The account name has not been supplied.");
                if (string.IsNullOrWhiteSpace(KeyValue))
                    throw new FormatException("The key value has not been supplied.");
                //NB: ForceBlobProperties is options, and defaults to false
            }
            catch (FormatException exception)
            {
                Console.WriteLine("{0}. Expected syntax is: {1}", exception.Message, SyntaxHelp);
                return false;
            }
            return true;
        }

        /// <summary>
        ///     Get the mime type of the file in the Uri
        /// </summary>
        private static string GetMimeType(Uri uri)
        {
            var fileExtension = Path.GetExtension(uri.ToString());
            //MimeTypes is an external dependency (nuget)
            return MimeTypeMap.GetMimeType(fileExtension);
        }

        /// <summary>
        ///     Return a Base64 string of the file specified
        /// </summary>
        private static string GetMd5(string filePath)
        {
            using (var md5 = MD5.Create())
            {
                using (var stream = File.OpenRead(filePath))
                {
                    return Convert.ToBase64String(md5.ComputeHash(stream));
                }
            }
        }

        /// <summary>
        ///     Recaptures a directory's name, to ensure the captialization is correct
        /// </summary>
        /// <param name="dirInfo">The directory's file path</param>
        /// <returns>The case sensitive file path</returns>
        private static string GetProperDirectoryCapitalization(DirectoryInfo dirInfo)
        {
            var parentDirInfo = dirInfo.Parent;
            if (null == parentDirInfo)
                return dirInfo.Name;
            return Path.Combine(GetProperDirectoryCapitalization(parentDirInfo),
                                parentDirInfo.GetDirectories(dirInfo.Name)[0].Name);
        }

        /// <summary>
        ///     Recaptures a file's full file path to ensure the captialization is correct
        /// </summary>
        /// <param name="filename">The file's file path</param>
        /// <returns>The case sensitive file path</returns>
        private static string GetProperFilePathCapitalization(string filename)
        {
            var fileInfo = new FileInfo(filename);
            var dirInfo = fileInfo.Directory;
            return (dirInfo == null)
                       ? null
                       : Path.Combine(GetProperDirectoryCapitalization(dirInfo),
                                      dirInfo.GetFiles(fileInfo.Name)[0].Name);
        }

        #endregion

        #region Properties

        private static string _localPath;
        private static string _containerName;
        private static string _accountName;
        private static string _keyValue;
        private static bool? _forceBlobProperties;

        /// <summary>
        ///     The local file path that we're syncing
        /// </summary>
        protected static string LocalPath
        {
            get
            {
                if (string.IsNullOrWhiteSpace(_localPath))
                    LocalPath = ConfigurationManager.AppSettings["LocalPath"];

                return _localPath;
            }
            set { _localPath = value; }
        }

        /// <summary>
        ///     The container name that we're syncing
        /// </summary>
        protected static string ContainerName
        {
            get
            {
                if (string.IsNullOrWhiteSpace(_containerName))
                    ContainerName = ConfigurationManager.AppSettings["ContainerName"];

                return _containerName;
            }
            set { _containerName = value; }
        }

        /// <summary>
        ///     The storage account name
        /// </summary>
        protected static string AccountName
        {
            get
            {
                if (string.IsNullOrWhiteSpace(_accountName))
                    AccountName = ConfigurationManager.AppSettings["AccountName"];

                return _accountName;
            }
            set { _accountName = value; }
        }

        /// <summary>
        ///     The storage account key value
        /// </summary>
        protected static string KeyValue
        {
            get
            {
                if (string.IsNullOrWhiteSpace(_keyValue))
                    KeyValue = ConfigurationManager.AppSettings["KeyValue"];

                return _keyValue;
            }
            set { _keyValue = value; }
        }

        /// <summary>
        ///     Force the blob properties to be overwritten even for existing files
        /// </summary>
        protected static bool ForceBlobProperties
        {
            get
            {
                if (!_forceBlobProperties.HasValue)
                {
                    bool b;
                    if (bool.TryParse(ConfigurationManager.AppSettings["ForceBlobProperties"], out b))
                        _forceBlobProperties = b;
                }

                return _forceBlobProperties ?? false;
            }
            set { _forceBlobProperties = value; }
        }

        #endregion
    }
}